//----------------------------------------------------------------------
// File:			ALTSfastLTS.cpp
// Programmer:		David Mount
// Description:		Search using the Fast-LTS heuristic
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include "ALTS.h"						// ALTS definitions
#include "ALTSfastLTS.h"				// Fast-LTS definitions
#include "ALTSnumeric.h"				// numeric utility definitions
#include "ALTSrand.h"					// random number generation

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	fastLTS
//	This is a partial implementation of the Fast-LTS heuristic of
//	Rousseuw and van Driessen. We perform a given number of random
//	elemental fits, and for each we perform a given number of
//	improvement steps. We keep the best 10 results, and on termination,
//	we apply improvement steps to these until convergence.
//	
//	Note that we maintain a vector of the best results. The vector
//	contains an additional sentinel element. Storing an element here
//	effectively deletes it. This just keeps the code cleaner.
//----------------------------------------------------------------------

LTSResult alts::fastLTS(
	const Params&		params)			// parameters
{
	ostream* out = params.out;			// output stream
										// number of best fits to keep
	int numBest = params.srchParams.numBest;
	int d = params.dim;					// dimension
	vector<LTSResult> best;				// storage for the best results
	best.reserve(numBest+1);			// reserve space for numBest
										// ...plus sentinel value
	for (int k = 0; k <= numBest; k++) {
		LTSResult r = LTSResult(d);		// add defaults contents
		r.setVariance(SCALAR_INFTY);
		best.push_back(r);
	}

	for (int i = 0; i < params.srchParams.numElemFits; i++) {
		bool doTrace = (params.traceFreq != 0 && i % params.traceFreq == 0);

		int n = (int) params.points.size(); // number of points
										// a random d-element subset
		vector<int> ranIndices = randomSubset(d, n);

		PointSet elem;					// the random elementary set
		for (int j = 0; j < d; j++) {	// insert d random points
			elem.insert(params.points[ranIndices[j]]);
		}

		Plane b(d);
										// ordinary least squares fit
		bool success = olsFit(elem, b);	// fit a plane to these points
		if (!success) {
			error("Failed to find fit in ltsFastLTS. Ignoring.", false);
			break;
		}
										// computes b's LTS variance
		Scalar bVariance = getLTSvariance(
					params.points,		// the points
					params.trimAbsValue,// trimming param
					b);					// existing plane
										// build result object
		LTSResult thisResult(bVariance, b, i);

		if (doTrace) {					// tracing?
			*out << "Sample: " << i
				 << " [";
			for (int j = 0; j < d; j++) *out << " " << ranIndices[j];
			*out << " ]\n  Base result: " << thisResult << "\n";
		}
										// apply improvement steps
		for (int j = 0; j < params.srchParams.numPivots; j++) {
			bool improved = thisResult.improveFit(params.points,
									params.trimAbsValue);
			if (doTrace) {				// tracing?
				*out << "  Improved: " << j
					 << " result = " << thisResult << "\n";
			}
			if (!improved) break;
		}
										// store among the best
		int k = numBest;
		Scalar thisVar = thisResult.getVariance();
										// slide up to make room
		while (k > 0 && (best[k-1].getVariance() > thisVar)) {
			best[k] = best[k-1];
			k--;
		}
		best[k] = thisResult;			// save this result
		if (doTrace) {					// tracing?
			*out << "  Stored: best[" << k << "]: " << thisVar
				 << " best[0]: " << best[0].getVariance() << endl;
		}
	}

	for (int k = 0; k < numBest; k++) {	// perform final improvements
										// something was stored here?
		if (best[k].getVariance() != SCALAR_INFTY) {
			LTSResult thisResult = best[k];
			if (params.traceFreq != 0) {
				*out << "Best[" << k
					 << "] = " << best[k].getVariance();
			}
			bool improved = true;
			int nSteps = 0;				// number of improvement steps
										// improve til convergence
			while (improved) {
										// locally improve
				improved = thisResult.improveFit(
						params.points,			// the points
						params.trimAbsValue);	// trimming param
				nSteps++;
			}
			best[k] = thisResult;		// save final result
			if (params.traceFreq != 0) {
				*out << " after " << nSteps << " steps: "
					 << "Final = " << best[k].getVariance() << endl;
			}
		}
	}

	int bestK = 0;						// get best of the best
	Scalar bestVariance = best[bestK].getVariance();
	for (int k = 1; k < numBest; k++) {
		if (best[k].getVariance() < bestVariance) {
			bestVariance = best[k].getVariance();
			bestK = k;
		}
	}
	return best[bestK];					// return best result
}
