//----------------------------------------------------------------------
// File:			ALTSmatrix.cpp
// Programmer:		David Mount
// Description:		Methods for matrix manipulation
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include <algorithm>					// for std::swap

#include "ALTSmatrix.h"					// definitions for matrices
#include "ALTSrand.h"					// random number generation

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	Utilities for vectors and matrices
//
//	uniformOnSphere(ctr, rad)	Returns uniform point on a sphere
//								centered at c with radius r.
//----------------------------------------------------------------------

Vect alts::uniformOnSphere(const Vect& ctr, Scalar rad)
{
	int d = ctr.getDim();				// dimension

	Vect r(d);							// result storage
	for (int j = 0; j < d; j++) {		// generate gaussian about origin
		r[j] = (Scalar) gauss(0, 1);
	}
	Scalar stretch = rad / r.length();	// stretch factor
	for (int j = 0; j < d; j++) {		// stretch and translate
		r[j] = (r[j] * stretch) + ctr[j];
	}
	return r;							// return result
}

//----------------------------------------------------------------------
//	Vect methods
//----------------------------------------------------------------------

							  			// construct from dim and base
Vect::Vect(int d, Scalar b) : vector<Scalar>(d)
{
	for (int j = 0; j < d; j++) (*this)[j] = b;
}

Vect::Vect(const Point& p)				// construct from Point
	: vector<Scalar>(p.getIndep().getDim())
{
	int d = getDim();
	(*this)[0] = p.getDep();			// (y, x1, ...., x{d-1})
	for (int j = 1; j < d; j++) {
		(*this)[j] = p.getIndep()[j];
	}
}

Scalar Vect::length() const				// return length
{
	int d = getDim();					// dimension
	Scalar ssq = 0;						// sum squared components
	const Vect& v = *this;
	for (int j = 0; j < d; j++) {
		ssq += v[j] * v[j];
	}
	return sqrt(ssq);
}

void Vect::normalize()					// normalize to unit length
{
	int d = getDim();					// dimension
	Vect& v = *this;
	Scalar len = length();				// get length
	for (int j = 0; j < d; j++) {
		v[j] /= len;
	}
}

void Vect::complement(const Vect& u)	// complement with respect to u
{
	int d = getDim();					// dimension
	Vect& v = *this;

	Scalar uDv = 0;						// holds <u.v>
	Scalar uDu = 0;						// holds <u.u>
	for (int j = 0; j < d; j++) {
		uDv += u[j] * v[j];
		uDu += u[j] * u[j];
	}
	Scalar stretch = uDv / uDu;			// stretch factor for projection
	for (int j = 0; j < d; j++) {
		v[j] -= stretch * u[j];
	}
}

Point Vect::castToPoint() const			// cast to Point
{
	int d = getDim();
	Scalar y = (*this)[0];				// (y, x1, ...., x{d-1})
	Indep x(d);
	for (int j = 1; j < d; j++) {
		x[j] = (*this)[j];
	}
	return Point(x, y);
}

//----------------------------------------------------------------------
//	Matrix methods
//----------------------------------------------------------------------

Matrix::Matrix(int d)					// construct dxd identity matrix
	: M(d)
{
	assert(d > 0);
	for (int i = 0; i < d; i++) {		// each row is a d-vector
		for (int j = 0; j < d; j++) {
			M[i].push_back(i == j ? 1 : 0);
		}
	}
}
										// return M * A
Matrix Matrix::operator*(const Matrix& A) const
{
	int d = getDim();					// dimension
	assert(d == A.getDim());			// check compatibility
	Matrix R(d);						// result matrix
	for (int i = 0; i < d; i++) {		// multiple A * M
		for (int j = 0; j < d; j++) {
			Scalar sum = 0;
			for (int k = 0; k < d; k++) {
				sum += M[i][k] * A[k][j];
			}
			R[i][j] = sum;
		}
	}
	return R;
}

//----------------------------------------------------------------------
//	randOrthog - set to a random orthogonal dxd matrix
//
//	This sets the matrix a random orthogonal dxd matrix.  The matrix is
//	generated by a quick and dirty (and probably not numerically sound)
//	method. We just apply Gram-Schmidt to a collection of vectors
//	sampled uniformly from a d-sphere.
//----------------------------------------------------------------------

void Matrix::randOrthog()
{
	const Scalar TOL = 1E-6;			// floating tolerance
	int d = getDim();					// dimension

	const Vect ORIGIN(d, 0);			// origin

	for (int i = 0; i < d; i++) {		// generate i-th row of M
		Vect u(d);
		do {							// try to get good vector
										// random vector on unit sphere
			u = uniformOnSphere(ORIGIN, 1);
										// apply Gram-Schmidt
			for (int j = 0; j < i; j++) {
				u.complement(M[j]);
			}
		} while (u.length() < TOL);		// too small? Try again
		u.normalize();					// normalize length
		M[i] = u;						// save in row i
	}
}

//----------------------------------------------------------------------
//	shiftAndCopy
//
//	Given a kxk matrix A, where k <= d and a shift amount s, where 0 <=
//	s <= d-k, this copies A[0..k-1][0..k-1] to M[s..s+k-1][s..s+k-1].
//----------------------------------------------------------------------

void Matrix::shiftAndCopy(Matrix A, int s)
{
	int d = getDim();					// dimension of M
	int k = A.getDim();					// dimension of K
	assert(0 <= s && s <= d-k);			// valid shift?

	for (int i = 0; i < k; i++) {		// copy A to M
		for (int j = 0; j < k; j++) {
			M[i+s][j+s] = A[i][j];
		}
	}
}

//----------------------------------------------------------------------
//	applyTo - apply matrix to various objects
//
//	This applies a matrix to various types of objects. The trickiest of
//	these is applying the matrix to a plane. We assume the plane vector
//	(b0, b1, ..., b{d-1}) corresponds to the plane:
//
//		y = b0 + b1*x1 + b2*x2 + ... + b{d-1}*x{d-1}
//
//	or equivalently
//
//		0 = b0 - y + b1*x1 + b2*x2 + ... + b{d-1}*x{d-1}
//
//	Let v = (1, y, x1, ..., x{d-1}) be the homogeneous (column) vector
//	representing the point (y, x), and let c be the (row) vector (b0,
//	-1, b1, ..., b{d-1}). Thus, we have c * v = 0.
//
//	Consider a matrix N formed by copying M into the lower-right dxd
//	submatrix of a (d+1)x(d+1) identity matrix. Applying M to the point
//	(y,x) is equivalent to computing N*v.
//
//	The desired result of applying M to b is, thus, to produce a
//	transformed plane b', whose corresponding homogeneous representation
//	c' satisfies c'*v' = 0.  From this we see that we want c'*(N*v) = 0.
//	Since c*v = 0, we see that it suffices to set c' = c*N^{-1}.
//
//	In general, we would need to invert N. We make the assumption (which
//	holds as of this release) that M is always an orthogonal matrix. It
//	follows that N is also orthogonal, and so N^{-1} = N^T (where "^T"
//	denotes transposition). Thus, we can compute (c')^T as N*(c^T). All
//	that remains is to convert this into the standard representation,
//	which we do by dividing all of c's components by the negation of its
//	2nd coordinate.
//----------------------------------------------------------------------

										// apply to plane
Plane Matrix::applyTo(const Plane& b) const	
{
	int d = getDim();					// dimension
	assert(d == b.getDim());			// check compatibility
	Plane rb(d);						// holds result
	rb[0] = b[0];
	Scalar sum = -1 * M[0][0];			// multiply M[0]*(-1,b1,..,b{d-1})
	for (int j = 1; j < d; j++) {
		sum += b[j] * M[0][j];
	}
	Scalar div = -sum;					// normalizing divisor

	for (int i = 1; i < d; i++) {
		sum = -1 * M[i][0];				// do M[1..d]*(-1,b1,..,b{d-1})
		for (int j = 1; j < d; j++) {
			sum += b[j] * M[i][j];
		}
		rb[i] = sum;
	}
	for (int j = 0; j < d; j++) {		// normalize
		rb[j] /= div;
	}
	return rb;
}

										// apply to Vect
Vect Matrix::applyTo(const Vect& v) const	
{
	int d = getDim();					// dimension
	assert(d == v.getDim());			// check compatibility
	Vect r(d);							// holds result
	for (int i = 0; i < d; i++) {
		Scalar sum = 0;
		for (int j = 0; j < d; j++) {
			sum += M[i][j] * v[j];
		}
		r[i] = sum;
	}
	return r;
}

										// apply to point
Point Matrix::applyTo(const Point& p) const	
{
	const Scalar& y = p.getDep();		// dependent and independent parts
	const Indep& x = p.getIndep();
	int d = getDim();					// dimension
	assert(d == x.getDim());			// check compatibility
	Scalar ry = 0;						// holds result
	Indep rx(d);
	for (int i = 0; i < d; i++) {
		Scalar sum = M[i][0] * y;
		for (int j = 1; j < d; j++) {
			sum += M[i][j] * x[j];
		}
		if (i == 0) ry = sum;			// store in result
		else rx[i] = sum;
	}
	return Point(rx, ry);
}

										// apply to vector of indep vars
Indep Matrix::applyTo(const Indep& x) const	
{
	int d = getDim();					// dimension
	assert(d == x.getDim());			// check compatibility
	Indep rx(d);						// holds result
	for (int i = 1; i < d; i++) {
		Scalar sum = 0;					// multiply as if y = 0
		for (int j = 1; j < d; j++) {
			sum += M[i][j] * x[j];
		}
		rx[i] = sum;
	}
	return rx;
}

//----------------------------------------------------------------------
//	I/O operators
//----------------------------------------------------------------------

										// output Vect
std::ostream& operator<<(std::ostream& out, const alts::Vect& v)
{
	int d = v.getDim();					// dimension
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "( ";
	for (int j = 0; j < d; j++) {
		out << v[j] << " ";
	}
	out << " )";
	return out;
}
										// output Matrix
std::ostream& operator<<(std::ostream& out, const alts::Matrix& M)
{
	int d = M.getDim();					// dimension
	out.precision(ALTS_OUTPUT_PRECISION);
	for (int i = 0; i < d; i++) {
		out << "[ ";
		for (int j = 0; j < d; j++) {
			out << M[i][j] << "\t";
		}
		out << "]\n";
	}
	return out;
}
