//----------------------------------------------------------------------
// File:			ALTSperf.cpp
// Programmer:		David Mount
// Description:		Performance measurement
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include "ALTSperf.h"					// ALTS perf definitions

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	PerfStat methods
//----------------------------------------------------------------------

void PerfStat::operator+=(Scalar x)		// add sample
{
	n++;
	sum += x;
	sumSq += x*x;
	if (x < minVal) minVal = x;
	if (x > maxVal) maxVal = x;
	dirty = true;
}

void PerfStat::update()					// update values
{
	meanVal = sum / n;
	varVal = (sumSq / n) - meanVal*meanVal;
	dirty = false;
}

Scalar PerfStat::mean()					// mean
{
	if (dirty) update();
	return meanVal;
}

Scalar PerfStat::variance()				// variance
{
	if (dirty) update();
	return varVal;
}

//----------------------------------------------------------------------
//	Performance statistics
//
//	The following data and routines are used for computing performance
//	statistics for nearest neighbor searching.  Because these routines
//	can slow the code down, they can be activated and deactiviated by
//	defining the PERF variable, by compiling with the option: -DPERF
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	Global counters for performance measurement
//	Most counters are indexed by level.
//----------------------------------------------------------------------

PerfCounts	alts::nProcCells;			// number of processed cells
PerfCounts	alts::nKilledCells;			// number of killed cells

										// level statistics (updated per cell)
static PerfStats lowerBnd;				// lower bounds
static PerfStats upperBnd;				// upper bounds
static PerfStats bestCost;				// best cost

										// global statistics (updated per cell)
static PerfStat	glLowerBnd;				// lower bounds
static PerfStat	glUpperBnd; 			// upper bounds
static PerfStat	glBestCost;				// best similarity

//----------------------------------------------------------------------
//	Routines for statistics.
//
//	The global nLevels stores the number of active levels. When a new
//	level is encountered, all the arrays are extended to this level.
//
//	For the sake of easier comparisons, we report the standard deviation
//	versions of lower and upper bounds.
//----------------------------------------------------------------------

static int nLevels = 0;					// number of active levels

static void resetUpToLevel(int lev)		// reset counts/stats up to level
{
	while (nLevels <= lev) {
		nProcCells.push_back(0);
		nKilledCells.push_back(0);
		lowerBnd.push_back(PerfStat());
		upperBnd.push_back(PerfStat());
		bestCost.push_back(PerfStat());
		nLevels++;
	}

	nProcCells[lev] = 0;				// explicitly reset level lev
	nKilledCells[lev] = 0;
	lowerBnd[lev].reset();
	upperBnd[lev].reset();
	bestCost[lev].reset();
}

void alts::resetLevelStats()			// reset counts/stats for each level
{
	nLevels = 0;
	glLowerBnd.reset();
	glUpperBnd.reset();
	glBestCost.reset();
}

void alts::updateCellStats(				// update stats for one cell
	const SlopeCell&	c,				// the cell
	const LTSResult&	best)			// best result so far
{
	int lev = c.level;
	if (c.level >= nLevels) {			// expand no. of levels if needed
		resetUpToLevel(lev);
	}
										// lower bound
	Scalar costLowerBnd = sqrt(c.lowerBnd);
	lowerBnd[lev]	+= costLowerBnd;
	glLowerBnd		+= costLowerBnd;

	if (c.status == ALIVE) {			// if alive get upper bound
		Scalar costUpperBnd = sqrt(c.upperBnd.getVariance());
		upperBnd[lev]	+= costUpperBnd;
		glUpperBnd		+= costUpperBnd;
	}

	Scalar bc = best.getCost();			// best cost so far
	if (bc < 1E10) {					// ignore outliers
		bestCost[lev]	 += bc;
		glBestCost		 += bc;
	}
}

										// add n to level lev
void PerfCounts::addToLevel(int lev, int n)
{
	if (nLevels <= lev) resetUpToLevel(lev);
	(*this)[lev] += n;
}

static void printOneStat(				// print a single statistic
	ostream&		out,				// output stream
	const string&	title,				// title
	PerfStat&		s,					// performance stat
	double			div)				// denominator
{
	out << title << "= [ ";
	out.width(9); out << s.mean()/div			<< " : ";
	out.width(9); out << s.stdDev()/div << " ]< ";
	out.width(9); out << s.min()/div			<< " , ";
	out.width(9); out << s.max()/div			<< " >\n";
}

void alts::printStats( 					// print statistics for a run
	std::ostream&		out,			// output stream
	bool				doLevStat)		// print level statistics?
{
	int glNProcCells = 0;				// count totals
	int glNKilledCells = 0;

	out.precision(ALTS_OUTPUT_PRECISION); // set floating precision
	out << "  (Performance stats:\n";
	for (int i = 0; i < nLevels; i++) {
		if (nProcCells[i] > 0) {
			if (doLevStat) {			// print level statistics
				out << "  level: " << i << "\n";
				out << "    processed-cells  = " << nProcCells[i]
					<< "  (killed = " << nKilledCells[i]
					<< " split = " << nProcCells[i] - nKilledCells[i]
					<< " )\n";
				out << "                       "
					<< "[      mean :    stddev ]<       min ,       max >\n";
				printOneStat(out, "    lower-bound      ", lowerBnd[i], 1);
				printOneStat(out, "    upper-bound      ", upperBnd[i], 1);
				printOneStat(out, "    best-cost        ", bestCost[i], 1);
			}
										// update totals
			glNProcCells		+= nProcCells[i];
			glNKilledCells		+= nKilledCells[i];
		}
	}
	out << "  all-levels:\n";			// print totals
	out << "    total-cells      = " << glNProcCells
		<< "  (killed = " << glNKilledCells
		<< " split = " << glNProcCells - glNKilledCells << " )\n";
	out << "                       "
		<< "[      mean :    stddev ]<       min ,       max >\n";
	printOneStat(out, "    lower-bound      ", glLowerBnd, 1);
	printOneStat(out, "    upper-bound      ", glUpperBnd, 1);
	printOneStat(out, "    best-cost        ", glBestCost, 1);
	out << "  )\n";
}
