//----------------------------------------------------------------------
// File:			ALTSprQueue.h
// Programmer:		David Mount
// Description:		Basic ALTS definitions
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#ifndef ALTS_PR_QUEUE_H
#define ALTS_PR_QUEUE_H

#include <iostream>						// C++ I/O
#include <vector>						// std::vector
#include <list>							// std::list

#include "ALTS.h"						// basic ALTS definitions
#include "ALTScell.h"					// ALTS slope cell definitions

//----------------------------------------------------------------------
//	Additions to alts namespace
//----------------------------------------------------------------------

namespace alts {
	class CellPrQueue;					// cell priority queue
	class CellPqNode;					// node in the queue
}

//----------------------------------------------------------------------
//	CellPqNode - Node in the priority queue
//
//	The priority queue is maintained using an STL priority queue.
//----------------------------------------------------------------------

class alts::CellPqNode {
public:
	CellPriorList	key;				// key value
	SlopeCell*		info;				// info field
public:
										// create from parts
	CellPqNode(CellPriorList k = 0, SlopeCell* i = NULL)
		: key(k), info(i) { }

										// component[i] comparator
	bool isLessThan(const CellPqNode& right, int i) const
	{ return key.isLessThan(right.key, i); }

										// lexicographical comparator
	bool operator<(const CellPqNode& right) const
	{ return key < right.key; }			// reversed for min-queue
};

//----------------------------------------------------------------------
//	CellPrQueue - Priority queue
//	A priority queue is a vector of items, along with associated
//	priorities. The basic operations are insert and extractMin.
//
//	Note that the STL's priority queue operation top() and pop()
//	return the maximum element in the queue. Thus, whenever minimization
//	is required, we store the negation of the priority in the queue.
//	Thus, the operator extractMin is improperly named, it really should
//	be called extractMax.
//
//	Each key of the queue is actually a vector of numeric keys. The
//	queue supports three possible access methods:
//
//	Lexicographic order: Extract elements in lexicographical order, so
//	that the first key of the vector has the highest priority.
//
//	Cyclic order: Select based on a cyclic permutation of keys, where
//	each successive extraction selects the minimum key according the
//	next component of the vector.
//
//	Weighted order: Select the key based on a weighted random selection.
//	In particular, each key is associated with positive weight. Prior to
//	the selection, the weights are normalized, so they sum to 1. We then
//	generate a random number, and determine the key according to the
//	probability distribution defined by the weights. If the extracted
//	cell is deemed to be a good choice, we "reward" this key by
//	increasing its weight, so it will be more likely used in the future.
//	If not, we may "penalize" the key by decreasing its weight.
//
//	We use two different storage methods, depending on which key
//	selection method is used. The lexicographic method uses a
//	heap-sorted vector and the other two methods use a linked list. We
//	store both structures, but when an attempt is made to switch from
//	one type to the other, we clear both structures, just to be safe.
//----------------------------------------------------------------------

class alts::CellPrQueue {
private:
	PQOrder order;						// ordering method
	int nKeys;							// number of keys
	int currKey;						// current key (for cyclic)
	std::vector<double> weights;		// weights (for weighted)
	std::vector<CellPqNode> vq;			// vector-based (lexicographic) queue
	std::list<CellPqNode> lq;			// list-based (cyclic order) queue

public:
										// iterator types
	typedef std::vector<CellPqNode>::const_iterator ConstVIterator;
	typedef std::list<CellPqNode>::const_iterator ConstLIterator;
	typedef std::list<CellPqNode>::iterator LIterator;

	CellPrQueue();						// default constructor
	bool usingLexOrder() const;			// using lexicographic order?
	bool usingCyclicOrder() const;		// using cyclic order?
	bool usingWeightedOrder() const;	// using weighted order?
	void clear();						// clear contents
	void useLexOrder();					// use lexicographic order
	void useCyclicOrder(int nk);		// use cyclic order
	void useWeightedOrder(int nk);		// use weighted order
	void nextCyclicKey();				// go to next key in cycle
	bool isEmpty() const;				// is queue empty?
	int size() const;					// size of queue
	void insert(SlopeCell*	inf);		// insert item into queue
	SlopeCell* extractMin();			// extract minimum
	Scalar getMinLowerBound() const;	// get minimum lower bound
	void normalizeWeights();			// normalize weights
	void rewardWeight(double score);	// reward current weight
	void printWeights(std::ostream* out) const; // print weight info
	void print(std::ostream* out) const; // print contents
};
#endif
