//----------------------------------------------------------------------
// File:			ALTSrand.cpp
// Programmer:		David Mount
// Description:		Methods for random data generation
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include <algorithm>					// for std::swap

#include "ALTSrand.h"					// definitions for generation
#include "ALTSmatrix.h"					// matrix manipulation definitions

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//------------------------------------------------------------------------
//	Methods for DistrParams
//------------------------------------------------------------------------

										// set to default values
void alts::DistrParams::setDefaults()
{
										// distribution
	distrib				= HYPERPLANE_UNIFORM;
	probInlier			=  1;			// probability of inlier
	for (int j = 1; j < dim; j++) {
		indepRange.lo[j] = -1;			// independent var range
		indepRange.hi[j] = +1;
		slopeRange.lo[j] = -1;			// slope range
		slopeRange.hi[j] = +1;
	}
	depRange.lo			= -1;			// dependent var range
	depRange.hi			=  1;
	interceptRange.lo	= -1;			// y-intercept range
	interceptRange.hi	=  1;
	residStdDev			=  0;			// residual standard deviation

										// for flat+sphere
	flatDim				=  1;			// flat dimension
	indepStdDev			=  0;			// residual standard deviation
	radius				=  1;			// radius of sphere
	probSpecial			=  0;			// prob. of a special point
}

//----------------------------------------------------------------------
//	Basic random number generators
//----------------------------------------------------------------------

static double ran0();					// uniform over [0,1]

//----------------------------------------------------------------------
//	Globals
//	idum is the seed.  Giving it a negative value resets the seed.
//
//	setSeed() sets a new seed value. getSeed() returns the current
//	seed value and forces the random number generator to reset
//	itself. This is done so that:
//
//			x = getSeed();
//			...generate random numbers
//
//			setSeed(x);
//			...generate random numbers
//
//		will generate the same random sequences.
//----------------------------------------------------------------------

static int		idum = 0;				// used for random number generation
static int		gauss_iset = 0;			// used for gauss generation

//------------------------------------------------------------------------
//	ran0 - (safer) uniform random number generator
//
//	The code given here is taken from "Numerical Recipes in C" by
//	William Press, Brian Flannery, Saul Teukolsky, and William
//	Vetterling. The task of the code is to do an additional randomizing
//	shuffle on the system-supplied random number generator to make it
//	safer to use. 
//
//	Returns a uniform deviate between 0.0 and 1.0 using the
//	system-supplied routine random() or rand(). Use setSeed()
//	to alter the random seed.
//------------------------------------------------------------------------

static double ran0()
{
	const int TAB_SIZE = 97;			// table size: any large number
	int j;

	static double y, v[TAB_SIZE];
	static int iff = 0;
	const double RAN_DIVISOR = double(RAND_MAX + 1UL);
	if (RAN_DIVISOR < 0) {
		std::cout << "RAN_DIVISOR " << RAN_DIVISOR << std::endl;
		exit(0);
	}

	//--------------------------------------------------------------------
	// As a precaution against misuse, we will always initialize on the
	// first call, even if "idum" is not set negative. Determine
	// "maxran", the next integer after the largest representable value
	// of type int. We assume this is a factor of 2 smaller than the
	// corresponding value of type unsigned int. 
	//--------------------------------------------------------------------

	if (idum <= 0 || iff == 0) {		// initialize
		iff = 1;
		ALTS_SRAND(idum);				// (re)seed the generator
		idum = 1;

		for (j = 0; j < TAB_SIZE; j++)	// exercise the system routine
			ALTS_RAND();				// (values intentionally ignored)

		for (j = 0; j < TAB_SIZE; j++)	// then save TAB_SIZE-1 values
			v[j] = ALTS_RAND();
		y = ALTS_RAND();				// generate starting value
	 }

	//--------------------------------------------------------------------
	// This is where we start if not initializing. Use the previously
	// saved random number y to get an index j between 1 and TAB_SIZE-1.
	// Then use the corresponding v[j] for both the next j and as the
	// output number.
	//--------------------------------------------------------------------

	j = int(TAB_SIZE * (y / RAN_DIVISOR));
	y = v[j];
	v[j] = ALTS_RAND();					// refill the table entry
	return y / RAN_DIVISOR;
}

//------------------------------------------------------------------------
//	ranUniform - generate a random uniform in [lo,hi]
//------------------------------------------------------------------------

double alts::ranUniform(
	double		lo,
	double		hi)
{
	return ran0()*(hi-lo) + lo;
}

//------------------------------------------------------------------------
//	randomInteger - generate a random integer from {lo, lo+1, ..., hi-1}
//
//	We assume that hi > lo.
//------------------------------------------------------------------------

int alts::randomInteger(
	int		lo,
	int		hi)
{
	int r = lo + (int) (ran0()*(hi-lo));
	if (r == hi) r--;					// (in case ran0() == 1)
	return r;
}

//------------------------------------------------------------------------
//	gauss - Gaussian random number generator
//	Returns a normally distributed deviate with zero mean and unit
//	variance, using ran0() as the source of uniform deviates.
//------------------------------------------------------------------------

double alts::gauss(
	double		mean,					// mean
	double		stdDev)					// standard deviation
{
	static double gset;

	if (gauss_iset == 0) {				// we don't have a deviate handy
		double v1, v2;
		double r = 2.0;
		while (r >= 1.0) {
			//------------------------------------------------------------
			// Pick two uniform numbers in the square extending from -1 to
			// +1 in each direction, see if they are in the circle of radius
			// 1.  If not, try again 
			//------------------------------------------------------------
			v1 = 2.0 * ran0() - 1.0; 
			v2 = 2.0 * ran0() - 1.0;
			r = v1 * v1 + v2 * v2;
		}
		double fac = sqrt(-2.0 * log(r) / r);
		//-----------------------------------------------------------------
		// Now make the Box-Muller transformation to get two normal
		// deviates.  Return one and save the other for next time.
		//-----------------------------------------------------------------
		gset = v1 * fac;
		gauss_iset = 1;					// set flag
		return (v2 * fac)*stdDev + mean;
	}
	else {								// we have an extra deviate handy
		gauss_iset = 0;					// so unset the flag
		return gset*stdDev + mean;		// and return it
	}
}

//----------------------------------------------------------------------
//	Distributions and Filters
//	A Distribution defines a probability distribution in the plane.
//	From this a number of concrete distribution classes are derived.
//	Each concrete class provides a constructor (which is particular to
//	the class) and is required to provide the following function:
//
//		Point genPoint()	returns a point from the distribution
//
//	A Filter defines a filter or sieve, which may be thought of as a
//	region of space.  If a point lies within this region it passes the
//	filter and other fails.	 Filter is an abstract class, and each
//	implementation of the class should implement the method
//
//		bool contains(p)	indicates whether p lies within
//							the filter region.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	Filter - a generic filter in the plane
//----------------------------------------------------------------------

class Filter {
public:
										// is p within the filter region?
	virtual bool contains(const Point& p) const = 0;
	virtual ~Filter() { }				// virtual destructor (C++ needs)
};

typedef Filter* FilterPtr;				// pointer to a filter

//----------------------------------------------------------------------
//	HalfplaneFilter
//	A closed halfplane lying to to the left side of the given line,
//	which is represented as a line segment.
//----------------------------------------------------------------------

class HalfplaneFilter : public Filter {
private:
	Plane	b;							// plane
	int		side;						// +1 for above and -1 for below
public:
										// default/standard constructor
	HalfplaneFilter(const Plane& bb, int s = +1)
	  { b = bb; side = s; }

	virtual bool contains(const Point& p) const
	  { return	side * b.residual(p) >= 0; }
};

//----------------------------------------------------------------------
//	Distribution - a generic distribution in the plane
//----------------------------------------------------------------------

class Distribution {
public:
	virtual Point genPoint() const = 0;	// sample a point
	virtual ~Distribution() { }			// virtual destructor (C++ needs)
};

typedef Distribution*	DistribPtr;		// pointer to a filter

//----------------------------------------------------------------------
//	UniformRect
//
//	Uniform in an axis-parallel hyper-rectangle
//----------------------------------------------------------------------

class UniformRect : public Distribution {
private:
	IndepRange	indepRange;				// ranges of independent vars
	ScalarRange	depRange;				// ranges of dependent vars
public:
										// constructor from ranges
	UniformRect(const IndepRange& ir, const ScalarRange& dr)
		: indepRange(ir), depRange(dr) { }

	virtual Point genPoint() const;		// sample a point
};

Point UniformRect::genPoint() const		// sample a point
{
	int d = indepRange.lo.getDim();		// dimension
	Scalar y = (Scalar) ranUniform(depRange.lo, depRange.hi);
	Indep x(d);
	for (int j = 1; j < d; j++) {
		x[j] = (Scalar) ranUniform(indepRange.lo[j], indepRange.hi[j]);
	}
	return Point(x, y);
}

//----------------------------------------------------------------------
//	UniformPlaneGauss
//
//	Points on a hyperplane with Gaussian noise added to the y
//	coordinate. The independent variables are distributed uniformly
//	along the independent range.
//----------------------------------------------------------------------

class UniformPlaneGauss : public Distribution {
private:
	IndepRange	indepRange;				// range of independent values
	Plane		b;						// the plane
	Scalar		stdDev;					// residual standard deviation
public:
										// default/basic constructor
	UniformPlaneGauss(
		const IndepRange& ir, 
		const Plane& bb, 
		Scalar sd = 0) : indepRange(ir), b(bb), stdDev(sd) { }

	virtual Point genPoint() const;		// sample a point
};

										// sample a point
Point UniformPlaneGauss::genPoint() const
{
	int d = indepRange.lo.getDim();		// dimension
	Scalar y = b[0];
	Indep x(d);
	for (int j = 1; j < d; j++) {
		x[j] = (Scalar) ranUniform(indepRange.lo[j], indepRange.hi[j]);
		y += b[j] * x[j];
	}
	y += gauss(0, stdDev);				// add noise
	return Point(x, y);
}

//----------------------------------------------------------------------
//	UniformSphere
//
//	Points distributed uniformly on a sphere centered at a given point
//	ctr.
//----------------------------------------------------------------------

class UniformSphere : public Distribution {
private:
	Vect	ctr;						// center
	Scalar	rad;						// radius
public:
										// default/basic constructor
	UniformSphere(const Vect& c, Scalar r = 1)
		: ctr(c), rad(r) { }

	virtual Point genPoint() const		// sample a point
		{ return uniformOnSphere(ctr, rad).castToPoint(); }
};

//----------------------------------------------------------------------
//	UniformEquator
//
//	Points distributed uniformly on an equatorial sphere at y = 0,
//	centered at a point ctr, and of the given radius. The points are
//	then subjected to a given transformation.
//----------------------------------------------------------------------

class UniformEquator : public Distribution {
private:
	Vect	ctr;						// center
	Scalar	rad;						// sphere radius
	Matrix	M;							// tilt transformation
	Scalar	residStdDev;				// residual standard deviation
public:
										// default/basic constructor
	UniformEquator(const Vect& c, const Matrix& MM, Scalar r = 1,
			Scalar rsd = 0) : ctr(c), rad(r), M(MM), residStdDev(rsd) { }

	virtual Point genPoint() const;		// sample a point
};


Point UniformEquator::genPoint() const
{
	int d = ctr.getDim();				// dimension
	Vect v(d);
	v[0] = 0;							// place on y=0 equatorial plane
	for (int j = 1; j < d; j++) {
		v[j] = (Scalar) gauss(0, 1);	// gauss is spherically symmetric
	}
	Scalar stretch = rad / v.length();	// stretch factor for desired radius
	for (int j = 0; j < d; j++) {		// stretch and translate
		v[j] = (v[j] * stretch) + ctr[j];
	}
										// add noise to dependent var
	v[0] += (Scalar) gauss(0, residStdDev);

	v = M.applyTo(v);					// transform v

	return v.castToPoint();				// cast to point and return
}

//----------------------------------------------------------------------
//	UniformDisk
//
//	Points distributed uniformly on an equatorial disk (a Gaussian cloud
//	on the plane y = 0 with a closed ball of a given radius).  The
//	points are then subjected to a given tilting transformation.
//----------------------------------------------------------------------

class UniformDisk : public Distribution {
private:
	Vect	ctr;						// center
	Scalar	rad;						// sphere radius
	Matrix	M;							// tilt transformation
	Scalar	residStdDev;				// residual standard deviation
public:
										// default/basic constructor
	UniformDisk(const Vect& c, const Matrix& MM, Scalar r = 1,
			Scalar rsd = 0) : ctr(c), rad(r), M(MM), residStdDev(rsd) { }

	virtual Point genPoint() const;		// sample a point
};


Point UniformDisk::genPoint() const
{
	int d = ctr.getDim();				// dimension
	Vect v(d);
	v[0] = 0;							// place on y=0 equatorial plane
	for (int j = 1; j < d; j++) {
		v[j] = (Scalar) gauss(0, 1);	// gauss is spherically symmetric
	}
	for (int j = 0; j < d; j++) {		// stretch and translate
		v[j] = (v[j] * rad) + ctr[j];
	}
										// add noise to dependent var
	v[0] += (Scalar) gauss(0, residStdDev);

	v = M.applyTo(v);					// transform v

	return v.castToPoint();				// cast to point and return
}

//----------------------------------------------------------------------
//	UniformFlatGauss
//
//	Independent variables are generated uniformly within the bounding
//	box along a f-dimensional flat with Gaussian noise (residStdDev in
//	the axis of dependent variables and indepStdDev in the remaining
//	d-f-1 axes of the independent variables). The resulting vector is
//	then subjected to a given transformation.
//
//	For technical reasons, the f-dimensional flat lies along dimensions
//	[d-f..d-1]. This is done to make it easier to define matrix M so
//	that it acts differently on these coordinates than the others.
//
//	Note: Since the random rotation occurs about the origin it is
//	recommended to have the independent range centered about the origin.
//----------------------------------------------------------------------

class UniformFlatGauss : public Distribution {
private:
	int				flatDim;			// dim of flat
	IndepRange		indepRange;			// range of independent values
	Scalar			residStdDev;		// residual standard deviation
	Scalar			indepStdDev;		// ind var standard deviation
	Matrix			M;					// transformation matrix
public:
										// construct from parts
	UniformFlatGauss(int fDim, const IndepRange& ir, const Matrix& MM, 
		Scalar rsd = 0, Scalar isd = 0)
		: flatDim(fDim), indepRange(ir), residStdDev(rsd),
		  indepStdDev(isd), M(MM) { }

	virtual Point genPoint() const;		// sample a point
};

										// sample a point
Point UniformFlatGauss::genPoint() const
{
	int d = indepRange.lo.getDim();		// dimension

	Vect v(d);
	v[0] += gauss(0, residStdDev);		// dep var is gaussian noise
	for (int j = 1; j < d; j++) {		// generate independent vars
		if (j < d-flatDim) {			// d-flatDim-1 are gaussian noise
			v[j] = (Scalar) gauss(0, indepStdDev);
		}
		else {							// coordinates on flat are uniform
			v[j] = (Scalar) ranUniform(indepRange.lo[j], indepRange.hi[j]);
		}
	}

	v = M.applyTo(v);					// transform v

	return v.castToPoint();				// convert to point and return
}

//------------------------------------------------------------------------
//	Mixed distribution point generators
//
//	These procedures handle the actual point generation. For each
//	distribution, we generate a vector of distribution generators and
//	filters, and invoke them to generate the actual distribution.
//------------------------------------------------------------------------

//------------------------------------------------------------------------
//	sampleMixture - sample points from distribution and filter
//
//	This procedure is given a number of points nPts, and two parallel
//	arrays: probs and distrs.  The first is a vector of probabilities
//	(which should sum to 1) and the second is a array of distribution
//	pointers. Based on probs, we determine how many points to generate
//	for each distribution.
//
//	We generate nPts. For each point we first select the distribution,
//	and then generate a point from this distribution.  If any filters
//	have been provided, we check that the point passes all the filters,
//	and if so it is copied to the final result vector.  When enough
//	points have been sampled, this vector is returned.
//
//	If the success rate (number of points that pass all the filters
//	divided by the total number of points generated) is less than
//	MIN_SUCCESS_RATE, then an error message is generated.
//------------------------------------------------------------------------

static void sampleMixture(
	int							nPts,	// number of points
	PointSet&					points,	// point set
	const vector<Scalar>&  		probs,	// probabilities
	const vector<DistribPtr>&	distrs,	// distributions
	const vector<FilterPtr>&	filters) // filters
{
	const Scalar MIN_SUCCESS_RATE = 0.01;// minimum success rate
	const Scalar MIN_ATTEMPTS = 100;	// minimum number of attempts

	int nDistr = (int) probs.size();	// number of distributions
	int nFilters = (int) filters.size();// number of filters
	if (nDistr == 0 || nDistr != int(distrs.size())) {
		error("Internal error 1 in sampleMixture", true);
	}

	vector<Scalar> accumProbs(nDistr);	// accumulated probabilities
	vector<int> accumSize(nDistr);		// accumulated sample sizes

	accumProbs[0] = probs[0];
	accumSize[0] = int(ceil(nPts * probs[0]));
	for (int i = 1; i < nDistr-1; i++){ // compute prefix sum
		accumProbs[i] = accumProbs[i-1] + probs[i];
		accumSize[i] = int(ceil(nPts * accumProbs[i]));
	}
	accumProbs[nDistr-1] = 1.0;			// force total to equal 1
	accumSize[nDistr-1] = nPts;

	int nSuccess = 0;					// number of successful samples
	int nAttempts = 0;					// number of attempts
	while (nSuccess < nPts) {			// repeat until getting enough

		int distrIdx = 0;
		while (accumSize[distrIdx] <= nSuccess) distrIdx++;

										// candidate sample point
		Point cand = distrs[distrIdx]->genPoint();
		nAttempts++;
										// try to pass all filters
		for (int j = 0; j < nFilters; j++) {
			bool successful = false;
			bool exhausted = false;
			while (!successful && !exhausted) {
				if (filters[j]->contains(cand)) {
					successful = true;	// we've passed this filter
				}
				else if (nAttempts > MIN_ATTEMPTS &&
					nAttempts * MIN_SUCCESS_RATE > nSuccess) {
					exhausted = true;	// too many tries
				}
				else {					// try again
					cand = distrs[distrIdx]->genPoint();
					nAttempts++;
				}
			}
			if (exhausted) {			// this is taking too long
				error("Failed to generate points after many attempts", true);
			}
		}
										// passed all filters
		//----------------DEBUG--------------------------------
		// (Output points with distribution labels)
		//-----------------------------------------------------
		// cout << "[" << nSuccess << "]: " << cand << " (" << distrIdx << ")" << endl;
		//-----------------------------------------------------
		points.push_back(cand);			// store successful candidate
		nSuccess++;
	}
}

//------------------------------------------------------------------------
//	randomHyperplane - generate a random hyperplane
//------------------------------------------------------------------------

static Plane randomHyperplane(
	const DistrParams&	distrParams)	// distribution parameters
{
	int d = distrParams.dim;			// dimension
	Plane thePlane(d);					// the hyperplane
										// hyperplane's y-intercept
	thePlane[0] = ranUniform(distrParams.interceptRange.lo,
						  distrParams.interceptRange.hi); 

	for (int j = 1; j < d; j++) {		// hyperplane's slope
		thePlane[j] = ranUniform(distrParams.slopeRange.lo[j],
						   distrParams.slopeRange.hi[j]); 
	}
	return thePlane;
}

//------------------------------------------------------------------------
//	genHyperplaneUniform - hyperplane + uniform outliers
//
//	This distribution involves inliers generated on a hyperplane with
//	gaussian residuals and outliers generated uniformly.
//------------------------------------------------------------------------

static Plane genHyperplaneUniform(
	int					nPts,			// number of points
	PointSet&			points,			// point set
	const DistrParams&	distrParams)	// distribution parameters
{
										// get random hyperplane
	Plane thePlane = randomHyperplane(distrParams);
	
	vector <DistribPtr> distribs(2);	// inlier/outlier distribs
	vector <Scalar> probs(2);			// inlier/outlier probs

	distribs[0] = new UniformPlaneGauss( // hyperplane + gauss error
			distrParams.indepRange, 
			thePlane, 
			distrParams.residStdDev);
	probs[0] = distrParams.probInlier;

										// uniform outliers
	distribs[1] = new UniformRect(distrParams.indepRange, distrParams.depRange);
	probs[1] = 1 - distrParams.probInlier;

	vector <FilterPtr> filters(0);		// no distribution filters
										// generate the points
	sampleMixture(nPts, points, probs, distribs, filters);

	delete distribs[0];					// delete allocated objects
	delete distribs[1];
	return thePlane;					// return estimated true fit
}

//------------------------------------------------------------------------
//	genHalfplaneUniform - generate data set from plane+half-unif
//
//	This distribution involves inliers generated on a hyperplane with
//	gaussian residuals and outliers generated uniformly, but filtered so
//	that they lie above the initial hyperplane.
//------------------------------------------------------------------------

static Plane genHalfplaneUniform(
	int					nPts,			// number of points
	PointSet&			points,			// point set
	const DistrParams&	distrParams)	// distribution parameters
{
										// get random hyperplane
	Plane thePlane = randomHyperplane(distrParams);
	vector <DistribPtr> distribs(2);	// inlier/outlier distribs
	vector <Scalar> probs(2);			// inlier/outlier probs
										// hyperplane + gauss error
	distribs[0] = new UniformPlaneGauss( // hyperplane + gauss error
			distrParams.indepRange, 
			thePlane, 
			distrParams.residStdDev);
	probs[0] = distrParams.probInlier;

										// uniform outliers
	distribs[1] = new UniformRect(distrParams.indepRange, distrParams.depRange);
	probs[1] = 1 - distrParams.probInlier;

	vector <FilterPtr> filters(1);		// one filter
										// halfplane filter
	filters[0] = new HalfplaneFilter(thePlane, +1);

										// generate the points
	sampleMixture(nPts, points, probs, distribs, filters);

	delete distribs[0];					// delete allocated objects
	delete distribs[1];
	delete filters[0];
	return thePlane;					// return estimated true fit
}

//------------------------------------------------------------------------
//	genFlatSphere - flat + uniform on sphere
//
//	This distribution is designed to cause problems for any algorithm
//	based on random sampling of elementary sets. It involves inliers
//	generated on an f-dimensional flat with Gaussian residuals and
//	outliers generated uniformly on a sphere of a given radius, which is
//	centered at the origin. In addition, a fraction of the outliers,
//	called "special points", are on the intersection of the sphere with
//	a (d-1)-dimensional hyperplane passing through the flat. We then
//	apply an orthogonal transformation that first randomly tilts the
//	equatorial plane while fixing the f-dimensional flat, and secondly
//	spins the independent variables about the axis of the dependent
//	variables.
//
//	To make this distribution truly nasty, make the special probability
//	very small, set the trimming quantile to (inlier probability +
//	special probability) and set the trimming quantile error to 0. This
//	forces the algorithm to use the special points. Next make the radius
//	huge relative to the independent range. This causes all the special
//	points to behave as leverage points, thus making the remaining set
//	of inliers virtually irrelevant. This thwarts any algorithm based
//	on random sampling, since the fit is essentially determined entirely
//	by the few equatorial points.
//
//	We generate the f-dimensional flat within a (d-1)-dimensional
//	hyperplane. To do this, we (conceptually) generate an f-dimensional
//	flat aligned with the last f coordinate axes, then we spin this flat
//	using an orthogonal transformation that acts on the
//	(d-1)-dimensional hyperplane spanned by the independent variables.
//	Finally, we tilt this hyperplane by applying an orthogonal
//	transformation in the full d-dimensional space.
//
//	To generate the tilting transformation, we create a random
//	orthogonal transformation in (d-f)-dimensional space, and then shift
//	it so that it applies only to the first d-f coordinates. To generate
//	the spinning transformation, we create a random orthogonal
//	transformation in (d-1)-dimensional space, and then shift it so that
//	it only applies to the independent variables. We then compose these
//	two matrices, first tilting and then spinning.
//
//	This transformation is applied to both the equatorial points and the
//	points on the flat. (There is no need to apply it to the points on
//	the sphere, since the sphere is invariant under any orthogonal
//	transformation.)
//------------------------------------------------------------------------

static Plane genFlatSphere(
	int					nPts,			// number of points
	PointSet&			points,			// point set
	const DistrParams&	distrParams)	// distribution parameters
{
	int d = distrParams.dim;			// dimension

	vector <DistribPtr> distribs(3);	// inlier/outlier distribs
	vector <Scalar> probs(3);			// inlier/outlier probs

	//--------------------------------------------------------------------
	// Apply tilt transformation
	//--------------------------------------------------------------------
	int f = distrParams.flatDim;		// dimension of flat
	Matrix tiltPart(d-f);				// partial tilt transformation
	tiltPart.randOrthog();
	Matrix tilt(d);						// full tilt transformation
	tilt.shiftAndCopy(tiltPart, 0);
	Matrix spinPart(d-1);				// partial spin transformation
	spinPart.randOrthog();
	Matrix spin(d);						// full spin transformation
	spin.shiftAndCopy(spinPart, 1);
	Matrix M(spin * tilt);				// final transformation
	//--------------------------------------------------------------------
	// Matrix M(d);						// use instead if no tilt desired
	//--------------------------------------------------------------------

										// uniform sphere outliers
	const Vect ORG(d, 0);				// origin of sphere and equator
										// sphere radius
	const Scalar& rad = distrParams.radius;
										// special points on tilted equator
	distribs[0] = new UniformEquator(ORG, M, rad, distrParams.residStdDev);
			
	probs[0] = distrParams.probSpecial;

	distribs[1] = new UniformFlatGauss( // flat + gauss error
			distrParams.flatDim,		// dimension of flat
			distrParams.indepRange, 	// independent range
			M,							// tilt and spin transformation
			distrParams.residStdDev,	// residual std dev
			distrParams.indepStdDev);	// indep var std dev
	probs[1] = distrParams.probInlier;

										// remaining outliers on sphere
	distribs[2] = new UniformSphere(ORG, distrParams.radius);

	probs[2] = 1 - distrParams.probInlier - distrParams.probSpecial;

	vector <FilterPtr> filters(0);		// no distribution filters
										// generate the points
	sampleMixture(nPts, points, probs, distribs, filters);

	delete distribs[0];					// delete allocated objects
	delete distribs[1];
	delete distribs[2];

	Plane thePlane(d, 1, 0);			// equatorial plane (y = 0)
	thePlane = M.applyTo(thePlane);		// transform via M
	return thePlane;					// return estimated true fit
}

//------------------------------------------------------------------------
//	genFlatRectangle - flat + uniform in hyper-rectangle
//
//	This distribution allows the user to control the degree of
//	conditioning in the distribution. The inliers are placed on an
//	f-dimensional flat F, which itself lies on a (d-1)-dimensional
//	hyperplane H. There are two types of noise added. The residStdDev
//	noise is added relative to the y-coordinate (and does not affect
//	conditioning) and the indepStdDev is added along H and does improve
//	conditioning. The outliers are uniform in a hyper-rectangle.  (See
//	the notes in the genFlatSphere for the generation of the tilting and
//	spinning transformations.)
//------------------------------------------------------------------------

static Plane genFlatRectangle(
	int					nPts,			// number of points
	PointSet&			points,			// point set
	const DistrParams&	distrParams)	// distribution parameters
{
	int d = distrParams.dim;			// dimension
	int f = distrParams.flatDim;		// dimension of flat

	vector <DistribPtr> distribs(2);	// inlier/outlier distribs
	vector <Scalar> probs(2);			// inlier/outlier probs

	Matrix tiltPart(d-f);				// partial tilt transformation
	tiltPart.randOrthog();
	Matrix tilt(d);						// full tilt transformation
	tilt.shiftAndCopy(tiltPart, 0);
	Matrix spinPart(d-1);				// partial spin transformation
	spinPart.randOrthog();
	Matrix spin(d);						// full spin transformation
	spin.shiftAndCopy(spinPart, 1);
	Matrix M(spin * tilt);				// final transformation

										// uniform sphere outliers
	const Vect ORG(d, 0);				// origin of sphere and equator

	distribs[0] = new UniformFlatGauss( // flat + gauss error
			distrParams.flatDim,		// dimension of flat
			distrParams.indepRange, 	// independent range
			M,							// tilt and spin transformation
			distrParams.residStdDev,	// residual std dev
			distrParams.indepStdDev);	// indep var std dev
	probs[0] = distrParams.probInlier;

										// remaining outliers on rectangle
	distribs[1] = new UniformRect(distrParams.indepRange, distrParams.depRange);
	probs[1] = 1 - distrParams.probInlier;

	vector <FilterPtr> filters(0);		// no distribution filters
										// generate the points
	sampleMixture(nPts, points, probs, distribs, filters);

	delete distribs[0];					// delete allocated objects
	delete distribs[1];

	Plane thePlane(d, 0, 0);			// equatorial plane (y = 0)
	thePlane = M.applyTo(thePlane);		// transform via M
	return thePlane;					// return estimated true fit
}

//------------------------------------------------------------------------
//	External entry points
//------------------------------------------------------------------------

//------------------------------------------------------------------------
//	generatePoints - generate points from distribution
//
//	This simply invokes the appropriate distribution-generation
//	procedure. It returns the estimated true plane of fit.
//------------------------------------------------------------------------

Plane alts::generatePoints(
	int					nPts,			// number of points
	PointSet&			points,			// point set
	const DistrParams&	distrParams)	// distribution parameters
{
	switch (distrParams.distrib) {		// switch on distribution type

	case HYPERPLANE_UNIFORM:			// hyperplane+unif
		return genHyperplaneUniform(nPts, points, distrParams);

	case HYPERPLANE_HALF_UNIFORM:		// hyperplane+half-unif
		return genHalfplaneUniform(nPts, points, distrParams);

	case FLAT_SPHERE:					// flat+sphere
		return genFlatSphere(nPts, points, distrParams);

	case FLAT_RECTANGLE:				// flat+rectangle
		return genFlatRectangle(nPts, points, distrParams);

	default:
		assert(false);
		return Plane();					// (to keep compiler happy)
	};
}

//----------------------------------------------------------------------
//	Seed setting
//	setSeed() sets a new seed value. getSeed() returns the current
//	seed value and forces the random number generator to reset
//	itself. This is done so that:
//
//			x = getSeed();
//			...generate random numbers
//
//			setSeed(x);
//			...generate random numbers
//
//	will generate the same random sequences.
//----------------------------------------------------------------------

void alts::setSeed(int seed)			// set seed value
{
	ALTS_SRAND(seed);					// set random's seed
	gauss_iset = 0;						// force ranGauss to reset
	idum = (seed < 0 ? seed : -seed);	// negate to force reset
}

int alts::getSeed()						// get seed value
{
	gauss_iset = 0;						// force ran_gauss to reset
	if (idum > 0) idum = -idum;			// force a reset
	return -idum;
}

//------------------------------------------------------------------------
//	randomSubset
//	Generates a random subset of m elements of {0..n-1} without
//	replacement. If m << sqrt(n), we just use rejection (which runs in
//	O(m^2) time in expectation). This simply retries whenever a
//	duplicate is generated. 
//------------------------------------------------------------------------

vector<int> alts::randomSubset(
	int 	m,
	int 	n)
{
	assert(m > 0 && m <= n);
	vector<int> set(m);
	if (m < int(sqrt(float(n)))) {		// if m small, use rejection
		for (int i = 0; i < m; i++) {
			int r;
			bool success = false;
			while (!success) {
				r = randomInteger(0, n);
				success = true;
				for (int j = 0; j < i; j++) {
					if (set[j] == r) {
						success = false;
						break;
					}
				}
			}
			set[i] = r;
		}
	}
	else {
		vector<int> v;					// if m large, use random swaps
		for (int i = 0; i < n; i++)
			v.push_back(i);
		int k = n;
		for (int j = 0; j < m; j++) {
			int r = randomInteger(0, k);
			swap(v[j], v[j+r]);
			k--;
		}
		for (int j = 0; j < m; j++) {
			set[j] = v[j];
		}
	} 
	return set;
}
