//----------------------------------------------------------------------
// File:			ALTSrand.h
// Programmer:		David Mount
// Description:		Definitions for random data generation
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#ifndef ALTS_RAND_H
#define ALTS_RAND_H

#include "ALTS.h"						// ALTS definitions

//----------------------------------------------------------------------
//	Although random/srandom is a more reliable random number generator,
//	many systems do not have it. If it is not available, set the
//	preprocessor symbol ALTS_NO_RANDOM, and this will substitute the use
//	of rand/srand for them.
//----------------------------------------------------------------------
#ifdef ALTS_NO_RANDOM					// for systems not having random()
  #define ALTS_RAND			rand
  #define ALTS_SRAND		srand
  #define ALTS_RAND_MAX		RAND_MAX
#else									// otherwise use rand()
  #define ALTS_RAND			random
  #define ALTS_SRAND		srandom
  #define ALTSM_RAND_MAX	2147483647UL	// 2**{31} - 1
#endif
//------------------------------------------------------------------------
//	Enumerations
//------------------------------------------------------------------------

namespace alts{

	enum Distrib {						// distributions
		HYPERPLANE_UNIFORM,				// hyperplane+unif
		HYPERPLANE_HALF_UNIFORM,		// hyperplane+half-unif
		FLAT_SPHERE,					// flat+sphere
		FLAT_RECTANGLE,					// flat+rectangle
		// ...
		N_DISTRIBS
	};

	class DistrParams;					// distribution parameters
}

//------------------------------------------------------------------------
//	DistrParams - distribution parameters
//	Used to store options related to generation of random point sets for 
//	testing purposes. Note that if the value of dim is changed, it is
//	necessary to allocate space for the vector objects (indepRange and
//	slopeRange).
//------------------------------------------------------------------------

class alts::DistrParams {
public:
	Distrib		distrib;				// distribution
	int			dim;					// dimension
	double		probInlier;				// probability of being an inlier
	IndepRange	indepRange;				// range of independent vars
	ScalarRange	depRange;				// range of dependent vars
	SlopeRange	slopeRange;				// range of slopes for data
	ScalarRange	interceptRange;			// range of y-intercepts
	Scalar		residStdDev;			// residual standard deviation
										// for flat+sphere
	int			flatDim;				// dimension of flat
	Scalar		indepStdDev;			// independent var std deviation
	Scalar		radius;					// radius of sphere with leverage pts
	Scalar		probSpecial;			// prob. of a special leverage pt
//
	void setDefaults();					// set to default values

	DistrParams(int d = 1)				// constructor
		: dim(d), indepRange(d), depRange(), slopeRange(d), interceptRange()
		{ setDefaults(); }
};

//------------------------------------------------------------------------
//	External entry points
//------------------------------------------------------------------------

namespace alts {
	void setSeed(int seed);				// set seed value

	int getSeed();						// get seed value

	double ranUniform(					// uniform over [lo,hi]
		double		lo,
		double		hi);

	int randomInteger(					// random int [lo,hi-1]
		int lo,
		int hi);

	double gauss(						// Gaussian random variable
		double	mean,					// with given mean and std_deviation
		double	stdDev);

	std::vector<int> randomSubset(		// random subset of k ints {0..n-1}
		int k,
		int n);

	Plane generatePoints(				// generate points
		int					nPts,			// number of points
		PointSet&			points,			// point set
		const DistrParams&	distrParams);	// distribution parameters
}

#endif
