//----------------------------------------------------------------------
// File:			ALTSsearch.cpp
// Programmer:		David Mount
// Description:		Main search procedure
// Last modified:	01/20/2016 (Version 1.1)
//----------------------------------------------------------------------
// Copyright (c) 2016 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//	Revision 1.1  01/20/2016
//		Changed B&B to ALTS
//----------------------------------------------------------------------

#include <algorithm>					// provides max

#include "ALTSsearch.h"					// ALTS search definitions
#include "ALTSutilities.h"				// ALTS utilities definitions
#include "ALTScell.h"					// ALTS slope cell definitions
#include "ALTSprQueue.h"				// ALTS priority queue definitions
#include "ALTSperf.h"					// ALTS performance stats

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	ltsSearch Utilities
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	initRoot - initialize root node for LTS search
//
//	Computes the slope samples and initial slope range for the root
//	node. There are two different ways to define the initial slope
//	range, first by specifying the slope_range explicitly, and second
//	by generating some number of samples, computing the bounding box
//	of these samples, and then applying an appropriate expansion factor
//	to the result. The final range is the smallest box that encloses
//	both sets of ranges.
//
//	When the bounding box of samples is used, we allow two additional
//	modifiers. First, we take a box that contains only a given fraction
//	of the sample points. Second, this box may be expanded by a given
//	scale factor.
//
//	Note that the root slope bounds in the search parameters is
//	modified by this procedure.
//----------------------------------------------------------------------

static void initRoot(
	SlopeCell* 			root,			// root cell (contents modified)
	Params&				params)			// parameters (slope bounds modified)
{
	int d = params.dim;					// dimension
										// search parameters
	SrchParams& srchParams = params.srchParams;
	SlopeRange& rootSlopeRange = srchParams.rootSlopeRange;
										// start with initial slope range
	rootSlopeRange = srchParams.initSlopeRange;
										// sampling requested?
	if (srchParams.slopeSampleSize > 0) {
		root->samples.generateSamples(params);
		if (srchParams.enableExpansion) { // expand range based on samples?
										// compute trimmed bounding box
			int h = int(srchParams.slopeSampleSize * srchParams.slopeTrimQuant);
			SlopeRange range = robustSlopeRange(d, root->samples, h);
										// scale the range
			range.expand(srchParams.slopeExpansion);
										// expand the range to contains samples
			for (int j = 1; j < d; j++) {
				if (range.lo[j] < rootSlopeRange.lo[j]) {
					rootSlopeRange.lo[j] = range.lo[j];
				}
				if (range.hi[j] > rootSlopeRange.hi[j]) {
					rootSlopeRange.hi[j] = range.hi[j];
				}
			}
		}
		else {							// crop samples to range
			cropSamplesToRange(root->samples, rootSlopeRange);
		}
	}
	root->slopeBnds = rootSlopeRange;	// store bounds with root
}

//----------------------------------------------------------------------
//	updateStatus
//
//	Computes the approximation error implied by the cell's lower bound.
//	If the error of the current best relative to the lower bound does
//	not exceed the approximate bounds, then the cell may be killed,
//	since any result this cell produces cannot contradict the
//	approximate optimality of the current best.
//----------------------------------------------------------------------

static void updateStatus(				// update cell status
	SlopeCell* 			cp,				// the cell (contents modified)
	const LTSResult&	best,			// best result
	const Params&		params)			// parameters
{
	Scalar bestCost = best.getCost();	// best cost so far
	Scalar lbCost = sqrt(cp->lowerBnd);
	Scalar relErr = (lbCost != 0 ? (bestCost - lbCost)/lbCost : SCALAR_INFTY);
	Scalar absErr = (bestCost  - lbCost);
										// can we kill this cell?
	if (relErr <= params.residRelErr || absErr <= params.residAbsErr) {
		cp->status = DEAD;
		KILL(cp->level, 1)				// one more killed
	}
}

//----------------------------------------------------------------------
//	procCell - process one cell
//
//	Computes the costs and priority associated with a cell and updates
//	its status. If the resulting upper bound is lower than the best so
//	far, we update the best. Note that when a cell is removed from the
//	priority queue, we update its status, since the best result may have
//	changed in the interim.  If the cell is alive after processing, it
//	is inserted in the priority queue, otherwise it is deleted.
//
//	If we are using weighted priority ordering, we reward the associated
//	priority order accordingly. We compute two scores. The first is the
//	ratio of the cell's lower bound to the best cost so far (the theory
//	being high lower bounds are desirable, since they allow cells to be
//	destroyed) and the second is the ratio of the best cost to the
//	cell's upper bound (the theory being that low upper bounds are
//	good).
//
//	For debugging purposes, we have two validity tests. First, we test
//	whether the lower-bound <= upper- bound. We only do this if the
//	upper-bound lies within the cell and if the trim error is 0, since
//	otherwise the upper-bound may be smaller. (E.g, Due to local
//	improvements or the fact that the quantile is smaller.) The second
//	is that the lower bound associated with this cell should not be
//	smaller than the lower bound of its parent (since it is a subcell of
//	the parent, and lower bound that applies to the parent must apply
//	the child as well).
//----------------------------------------------------------------------
//	Bug: In the current implementation, if traceFreq > 0, then tracing
//	is done on *all* cells, not just periodically. I cannot recall why
//	I set it up this way, presumably to deal with a bug. I should fix
//	this.
//----------------------------------------------------------------------

static void procCell(					// compute cell costs and status
	SlopeCell* 			cp,				// the cell (contents modified)
	SlopeCell* 			par,			// the cell's parent
	LTSResult&			best,			// best result (modified)
	const Params&		params,			// parameters
	CellPrQueue&		pq)				// priority queue
{
	Scalar bestVar = best.getVariance();// best cost so far
	cp->getLowerBound(params);			// update/get lower bound
	if (pq.usingWeightedOrder()) {		// using weighted queue order?
										// lower-bound score = lower/best
		Scalar lbScore = sqrt(cp->lowerBnd / bestVar);
		pq.rewardWeight(lbScore);		// reward priority queue weight
	}
	updateStatus(cp, best, params);		// determine if cell dies
	if (cp->status != DEAD) {			// still living
		cp->getUpperBound(params);		// get upper bound
		if (pq.usingWeightedOrder()) {	// using weighted queue order?
			Scalar ubScore =			// upper-bound score = best/upper
				sqrt(bestVar / cp->upperBnd.getVariance());
			pq.rewardWeight(ubScore);	// reward priority queue weight
		}
										// better than current best?
		if (cp->upperBnd.getVariance() < bestVar) {
			best = cp->upperBnd;		// yes, new best
		}
		cp->setPriorities(params);		// set cell's priority
										// sanity check! (upper < lower?)
		if (cp->upperBnd.getVariance() < cp->lowerBnd - SCALAR_EPS &&
			params.trimAbsErr == 0 &&
			cp->slopeBnds.contains(cp->upperBnd.getPlane().getSlope())) {
			cp->trace(best, params);	// trace this cell
			alts::error("Internal error: upper < lower", true);
		}
										// sanity check! (lower < par.lower?)
		if (par != NULL && cp->lowerBnd < par->lowerBnd - SCALAR_EPS) {
			cp->trace(best, params);	// trace this cell
			cerr << "Lower = " << cp->lowerBnd << " Parent = " << par->lowerBnd << "\n";
			*params.out << "Lower = " << cp->lowerBnd << " Parent = " << par->lowerBnd << "\n";
			alts::error("Internal error: lower < par.lower", false);
		}
	}

	if (cp->status != DEAD) {			// still alive?
		pq.insert(cp);					// add to priority queue
	}
	if (params.traceFreq != 0) {		// tracing desired?
		cp->trace(best, params);		// trace this cell
	}
	#ifdef PERF							// performance evaluation
		PROC(cp->level, 1)				// one more processed
		alts::updateCellStats(*cp, best); // update cell statistics
	#endif
	if (cp->status == DEAD) delete cp;	// kill if dead
}

//----------------------------------------------------------------------
//	ltsSearch - main entry point for the adaptive-LTS search procedure
//
//	The search operates by inserting the initial cell into the priority
//	queue, and until the termination conditions are met, the algorithm
//	extracts a cell from the queue, processes the cell, and if the cell
//	is not killed, it then splits the cell and enqueues both on the
//	priority queue.
//
//	The search loop continues until either:
//		- the queue of cells is empty (meaning that all cells have been
//		  killed)
//		- the best similarity has fallen below the absolute error
//		- maxCells have been expanded.
//
//	This procedure is responsible for deleting all cells once they have
//	been killed.
//
//	The root slope bounds in the search parameters is modified
//	(indirectly through initRoot).
//----------------------------------------------------------------------

LTSResult alts::ltsSearch(
	Params&				params)			// parameters (modified)
{
	int d = params.dim;					// dimension
	#ifdef PERF							// performance evaluation
		alts::resetLevelStats();		// reset level-based stats
	#endif

	int cellCount = 0;					// initialize cell counter
	int maxCells = params.srchParams.maxCells;
	LTSResult best(d);					// stores best result
	best.setVariance(SCALAR_INFTY);
	CellPrQueue pq;						// priority queue
	switch(params.srchParams.pqOrder) {	// set priority type
		case LEX_ORDER:
			pq.useLexOrder();			// lexicographical
			break;
		case CYCLIC_ORDER:				// cyclic order
			pq.useCyclicOrder(params.srchParams.srchPriorList.size());
			break;
		case WEIGHTED_ORDER:			// weighted random order
			pq.useWeightedOrder(params.srchParams.srchPriorList.size());
			break;
	}

	SlopeCell* root = new SlopeCell(d);	// create root cell
	initRoot(root, params);				// initialize the root cell
	root->cellNum = ++cellCount;		// set cell number
	procCell(root, NULL, best, params, pq);	// process cell

	do {								// main loop
		if (params.traceFreq != 0) {	// tracing desired?
			pq.print(params.out);		// print priority queue contents
										// print queue's min lower bound
			*params.out << "  pq-min-lb:   " << pq.getMinLowerBound() << "\n";
		}
		SlopeCell* cp = pq.extractMin();// extract current cell
		if (params.traceFreq != 0) {	// tracing desired?
			*params.out << "Extracting cell " << cp->cellNum << "\n";
		}
		updateStatus(cp, best, params);	// update status (for current best)

		if (cp->status != DEAD) {		// cell is still alive?
			SlopeCell* c1 = new SlopeCell(d);
			SlopeCell* c2 = new SlopeCell(d);
			c1->cellNum = ++cellCount;	// assign cell numbers
			c2->cellNum = ++cellCount;

			cp->split(*c1, *c2, params);// split the cell
										// process the cells
			procCell(c1, cp, best, params, pq);
			procCell(c2, cp, best, params, pq);
		}
		delete cp;						// delete parent cell

	} while (!(pq.isEmpty() ||			// stop when queue is empty
										// ...or too many cells
		(maxCells != 0 && cellCount > maxCells) ||
										// ...or cost is low enough
		best.getCost() < params.residAbsErr));

	return best;						// return best result
}
