//----------------------------------------------------------------------
// File:			ALTSslopeSet.cpp
// Programmer:		David Mount
// Description:		Methods for slope sets
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include "ALTS.h"						// for slope set definitions
#include "ALTSnumeric.h"				// numeric utility definitions
#include "ALTSrand.h"					// random number generation

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	generateSamples - Generates slope samples by random elemental fits.
//----------------------------------------------------------------------

void SlopeSet::generateSamples(
	const Params&		params)			// parameters
{
	ostream* out = params.out;			// output stream
	int d = params.dim;					// dimension
										// number of samples
	int numSample = params.srchParams.slopeSampleSize;

	for (int i = 0; i < numSample; i++) {
		int n = (int) params.points.size(); // number of points
										// a random d-element subset
		vector<int> ranIndices = randomSubset(d, n);

		PointSet elem;					// the random elementary set
		for (int j = 0; j < d; j++) {	// insert d random points
			elem.insert(params.points[ranIndices[j]]);
		}

		Plane b(d);
										// ordinary least squares fit
		bool success = olsFit(elem, b);	// fit a plane to these points
		if (!success) {
			error("Failed to find fit in SlopeSet::sample. Ignoring.", false);
			break;
		}
		insert(b.getSlope());			// add to slope samples
		if (params.traceFreq != 0) {	// tracing?
			*out << "Sample: " << i
				 << " [";
			for (int j = 0; j < d; j++) *out << " " << ranIndices[j];
			*out << " ]: ";
			*out << "Slope = " << b.getSlope() << "\n";
		}
	}
}

//----------------------------------------------------------------------
//	getBoundingBox - Get bounding box for slope set
//----------------------------------------------------------------------

SlopeRange SlopeSet::getBoundingBox() const
{
	assert(size() > 0);					// must be nonempty
	SlopeSetConstIter it = begin();
	int d = it->getDim();				// dimension
	SlopeRange range(Slope(d, +SCALAR_INFTY), Slope(d, -SCALAR_INFTY));
	while (it != end()) {
		for (int j = 1; j < d; j++) {
			if ((*it)[j] < range.lo[j]) {	// new minimum?
				range.lo[j] = (*it)[j];
			}
			if ((*it)[j] > range.hi[j]) {	// new maximum?
				range.hi[j] = (*it)[j];
			}
		}
		it++;
	}
	return range;
}

//----------------------------------------------------------------------
//	getVariance - Get the spatial variance of a slope set
//
//	The variance of a slope set is used as a cell priority. Intuitively,
//	if slope samples of a cell are densely clustered around a single
//	slope, this would suggest that there are many witnesses within this
//	cell of a good fit.
//
//	In the code below, we apply the formula:
//
//               1
//		var = ------ sum_{i=1}^n (X[i])^2 - (sum_{i=1}^n X[i])^2
//            n(n-1)
//
//	where n is the number of slopes, X[i] is the ith slope, and squaring
//	means dot product of a vector with itself.
//----------------------------------------------------------------------

Scalar SlopeSet::getVariance() const
{
	int n = size();						// number of slopes
	if (n <= 1) return 0;				// 0 or 1 slopes? Zero variance
	SlopeSetConstIter it = begin();
	int d = it->getDim();				// dimension
	Slope vSum(d, 0);					// vector sum of slopes
	Scalar sumSq = 0;					// sum of squared slopes
	while (it != end()) {
		Scalar thisSq = 0;				// current vector squared
		for (int j = 1; j < d; j++) {
			vSum[j] += (*it)[j];
			thisSq += (*it)[j] * (*it)[j];
		}
		sumSq += thisSq;				// add current square to sum
		it++;
	}
	Scalar sqVSum = 0;					// square of vSum
	for (int j = 1; j < d; j++) {
		sqVSum += vSum[j] * vSum[j];
	}
										// compute final variance
	Scalar variance = (sumSq - (sqVSum/n))/(n-1);
	return variance;
}

//----------------------------------------------------------------------
//	partition - Partition a slope set about a given cut value.
//
//	Basically, we partition points about the cut value. However, some
//	care is needed to handle points that lie on the cutting plane, so
//	that the overall set is split as evenly as possible. We do this as
//	follows. We count the number of values that are less than the
//	cutting value (nLT) and the numer equal (nEq). We partition the
//	equal elements in order to produce the most balanced split possible.
//----------------------------------------------------------------------

void SlopeSet::partition(
	int			cutDim,					// cutting dimension
	Scalar		cutVal,					// cutting value
	SlopeSet&	lowSide,				// slopes on the low side
	SlopeSet&	highSide) const			// slopes on the high side
{
	lowSide.clear();					// clear out sets
	highSide.clear();
	if (size() == 0) return;			// nothing to partition

	int nLT = 0;						// counts of less-than and equal
	int nEq = 0;
	SlopeSet midSide;					// points on the cutting plane
	SlopeSetConstIter it = begin();		// iterator
	while (it != end()) {
		if ((*it)[cutDim] < cutVal) {	// low side
			lowSide.push_back(*it);
			nLT++;
		}
		else if ((*it)[cutDim] > cutVal) { // high side
			highSide.push_back(*it);
		}
		else {							// on the splitting plane
			midSide.push_back(*it);
			nEq++;
		}
		it++;
	}

	int median = int(size())/2;			// median index
	int nLo = median - nLT;				// how many mids to add to low
	if (nLo < 0) nLo = 0;
	else if (nLo > nEq) nLo = nEq;

	for (int i = 0; i < nEq; i++) {		// copy mids to low/high sides
		if (i < nLo) lowSide.push_back(midSide[i]);
		else highSide.push_back(midSide[i]);
	}
}
