//----------------------------------------------------------------------
// File:			ALTSutilities.cpp
// Programmer:		David Mount
// Description:		Interval LTS
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include <algorithm>					// for sort()

#include "ALTSutilities.h"				// LTS utilities definitions
#include "ALTSbasicStat.h"				// basic stat definitions

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	cropSamplesToRange - crop samples to range
//----------------------------------------------------------------------

void alts::cropSamplesToRange(
	SlopeSet& 			samples, 		// the slope samples (modified)
	const SlopeRange& 	slopeRange)		// slope range
{
	int nSamp = (int) samples.size();	// number of samples
	SlopeSet cropped;					// cropped samples
	for (int i = 0; i < nSamp; i++) {	// check each sample
		if (slopeRange.contains(samples[i])) {
			cropped.push_back(samples[i]); // if inside, keep it
		}
	}
	samples = cropped;					// replace samples with cropped
}

//----------------------------------------------------------------------
//	robustSlopeRange - robust bounding box for slope samples
//
//	This is a quick and dirty heuristic for the following robust
//	estimation problem. Given a finite set of n slope samples (each a
//	point in (d-1)-dimensional space) and a trimming parameter h,
//	compute the smallest volume hyperrectangle that encloses h of these
//	points. The approach is to apply the 1-dimensional LMS algorithm to
//	trim away points according to their projection along each coordinate
//	axis.
//
//	More formally, define x = (h/n)^{1/(d-1)}. We prune the points
//	according to their projection along each of the axes, keeping a
//	fraction of x of the remaining points based on their projection
//	along the j-th coordinate axis, for 1 <= j <= d-1. After repeating
//	this d-1 times, a fraction of x^{d-1} = h/n points remain.
//
//	In order to prune the points along each axis, we sort the points
//	according to their projection along the j-th axis using sort(), and
//	then compute the interval of the minimum width containing a fraction
//	of x of points.  We then remove the points lying outside this
//	interval.
//
//	Tech note: In order to apply the STL sort procedure to individual
//	coordinates, we adopt a sneaky trick. We maintain a global variable,
//	activeDim, that holds the current coordinate index, which the
//	comparison procedure uses.
//----------------------------------------------------------------------

static int activeDim = 0;				// active sorting dimension
										// compare active coordinate
static bool coordLessThan(const Slope& s1, const Slope& s2)
{ return s1[activeDim] < s2[activeDim]; }

SlopeRange alts::robustSlopeRange(
	int 				d,				// dimension
	SlopeSet& 			samples, 		// the slope samples (modified)
	int					h)				// trimming value
{
	if (h == 0) {						// nothing to remove?
		return samples.getBoundingBox();// return bounding box
	}
	int nSamp = (int) samples.size();	// total number of points
	double pruneFraction = pow(double(h)/double(nSamp), 1/double(d-1));

										// prune w.r.t. each coord
	for (int j = 1; j < d; j++) {
										// number of points to keep
		int thisKeep = int(ceil(nSamp * pruneFraction));
		if (thisKeep < h) {
			thisKeep = h;				// ensure that at least h remain
		}
										// ensure we have h at the end
		if (j == d-1 && thisKeep != h) {
			thisKeep = h;
		}
		activeDim = j;					// set active dimension
										// sort based on this dimension
		sort(samples.begin(), samples.end(), coordLessThan);

		IntRange best(0, thisKeep);		// best interval storage
		Scalar bestWidth = samples[thisKeep-1][j] - samples[0][j];
										// process each interval
		for (int i = 1; i <= nSamp-thisKeep; i++) {
										// get current width
			Scalar width = samples[i+thisKeep-1][j] - samples[i][j];
			if (width < bestWidth) {	// best so far?
				bestWidth = width;
				best = IntRange(i, i+thisKeep);
			}
		}
										// erase high outliers
		samples.erase(samples.begin()+best.hi, samples.end());
										// erase low outliers
		samples.erase(samples.begin(), samples.begin()+best.lo);

		nSamp = thisKeep;				// update number of samples
	}
	return samples.getBoundingBox();		// bounding box of samples
}

//----------------------------------------------------------------------
//	scalarLTS - 1-dimensional (point) LTS problem
//
//	This solves the 1-dimensional LTS problem for points. It uses a 
//	standard sweep procedure, by considering intervals each containing
//	h points, and incrementally updating the cost for each.
//----------------------------------------------------------------------

ScalarLTSResult alts::scalarLTS(
	const ScalarSet& 	points, 		// the 1-dim points
	int					h)				// trimming value
{
	int nPts = (int) points.size();		// number of points

	vector<Scalar> pts(nPts);			// storage for sorted points

	for (int i = 0; i < nPts; i++) {	// copy points to pts
		pts[i] = points[i];
	}

	sort(pts.begin(), pts.end());		// sort them

	int lo = 0;							// initial interval [0..h-1]
	int hi = h-1;

	BasicStat stat;						// statistics for active sequence
	for (int i = lo; i <= hi; i++) {	// add contributions to init interval
		stat += pts[i];					// init stats for active sequence
	}
										// stores best result so far
	ScalarLTSResult best(SCALAR_INFTY, SCALAR_INFTY);
	for (int i = 0; i <= nPts-h; i++) {	// process each active sequence
		Scalar loc = stat.mean();		// get best location based on stats
		Scalar variance = stat.variance(); // get variance
		assert(variance >= 0);			// sanity check
		if (variance < best.variance) {	// best so far?
			best = ScalarLTSResult(loc, variance);
		}
										// advance to next active sequence
		stat -= pts[lo];				// remove contribution of old lo
		lo++;							// advance lo
		if (++hi < nPts) {				// advance hi
			stat += pts[hi];			// add contribution of new hi
		}
	}
	return best;
}

//----------------------------------------------------------------------
//	Interval Endpoint	
//----------------------------------------------------------------------

enum Side {								// interval side
	LO,
	HI };

class EndPt {
public:
	Scalar		b;						// endpoint
	Side		side;					// interval side
public:
										// constructor
	EndPt(Scalar bb = 0, Side s = LO) : b(bb), side(s) { }
};

										// comparison for sort()
bool operator<(const EndPt& e1, const EndPt& e2)
{ return e1.b < e2.b; }

										// output endpoint
ostream& operator<<(ostream& out, const EndPt& e)
{
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "( " << e.b << " " << (e.side == LO ? "LO" : "HI") << ")";
	return out;
}

//----------------------------------------------------------------------
//	intvLTS - 1-dimensional interval LTS problem
//
//	Given a set of intervals and a trimming parameter h, this procedure
//	computes the point that minimizes the sum of squared deviates to the
//	h closest points, one point within each interval.
//
//	The algorithm begins by sorting the intervals into a list of
//	endpoints. We then traverse this list maintaining three pointers,
//	lo, hi, and mid. The interval [lo..hi] marks the boundary of a
//	minimal subsequence that intersects h consecutive intervals. The lo
//	value points to the HI end of the first interval of the subsequence,
//	and the hi value points to the LO end of the last interval. Call
//	this the "active sequence."
//
//	The value mid indicates a elementary subinterval (that is, an
//	interval between two consecutive endpoints) I = [mid..mid+1]. Each
//	interval in the active sequence that lies entirely to the left of I
//	contributes its HI endpoint to the cost, each that lies entirely to
//	the right of I contributes its LO endpoint, and the others
//	contribute 0.
//
//	Because we are minimizing the sum of squared deviates, it follows
//	that the mean of the current sample is the optimum location.  We
//	compute the mean value based on the current contributions.  If it
//	lies to the right of I, we advance mid. If it lies within this
//	interval, we update the best so far and advance lo and hi to the
//	next endpoints of the appropriate variety.
//
//	Because intervals that contain the current elementary subinterval 
//	do not contribute to the mean, they have the effect of biasing the
//	variance (since they are not included in the sample statistics and
//	so artificially decrease the sample size, which should be h). If we
//	let m be the sample size, we correct this by adjusting the variance
//	by multiplying by m-1 and dividing by h-1.
//----------------------------------------------------------------------

ScalarLTSResult alts::intvLTS(
	const ScalarRangeSet& 	intervals, 	// the intervals
	int						h)			// trimming value
{
	int nIntv = (int) intervals.size();	// number of intervals
	vector<EndPt> endPt(2*nIntv);		// create a vector of endpoints

	for (int i = 0; i < nIntv; i++) {
		endPt[2*i] = EndPt(intervals[i].lo, LO);
		endPt[2*i + 1] = EndPt(intervals[i].hi, HI);
	}
	int nEnds = (int) endPt.size();		// number of endpoints

	sort(endPt.begin(), endPt.end());	// sort the endpoints

	int lo = -1;						// init lo is first HI endpoint
	do { lo++; } while (lo < nEnds && endPt[lo].side != HI);

	int hi = -1;
	for (int i = 0; i < h; i++) {		// init hi is h-th LO endpoint
		do { hi++; } while (hi < nEnds && endPt[hi].side != LO);
	}

	BasicStat stat;						// statistics for active sequence
	stat += endPt[lo].b;				// init stats for active sequence
	for (int i = lo+1; i <= hi; i++) {	// ...only LO sides contribute
		if (endPt[i].side == LO) stat += endPt[i].b;
	}
										// stores best result so far
	ScalarLTSResult best(SCALAR_INFTY, SCALAR_INFTY);
	int mid = lo;						// starting elementary interval

	for (int i = 0; i <= nIntv-h; i++) {// process each active sequence
		if (endPt[lo].b >= endPt[hi].b) {
										// h intervals overlap - cost = 0
			best = ScalarLTSResult(endPt[lo].b, 0);
			break;
		}
		Scalar loc = stat.mean();		// get best location based on stats
		assert(loc >= endPt[mid].b);	// loc cannot be behind us
		while (loc > endPt[mid+1].b) {	// opt loc is still ahead?
			mid++;						// advance to next elem interval
			switch (endPt[mid].side) {	// contribution of current interval
				case LO: stat -= endPt[mid].b; break;
				case HI: stat += endPt[mid].b; break;
			}
			loc = stat.mean();			// update location
			assert(loc >= endPt[mid].b); // cannot be behind us
		}
										// local opt in this elem interval
		int m = stat.nSamples();		// number of contributing intervals
										// get (adjusted) variance	
#ifdef ALTS_USE_BIASED_VARIANCE
		Scalar variance = stat.variance() * double(m) / double(h);
#else
		Scalar variance = stat.variance() * double(m-1) / double(h-1);
#endif
		if (variance < best.variance) {	// best so far?
			best = ScalarLTSResult(loc, variance);
		}
										// advance to next active sequence
		stat -= endPt[lo].b;			// remove contribution of old lo
										// advance lo
		do { lo++; } while (lo < nEnds && endPt[lo].side != HI);
										// advance hi
		do { hi++; } while (hi < nEnds && endPt[hi].side != LO);
		if (hi < nEnds) 
			stat += endPt[hi].b;		// add contribution of new hi
	}
	return best;
}
